package ida.pelab.cgn.ccode;

import java.util.ArrayList;


import cetus.hir.ArrayAccess;
import cetus.hir.CompoundStatement;
import cetus.hir.Program;
import cetus.hir.Traversable;

import ida.pelab.cgn.ICodeGenerator;
import ida.pelab.prt.annotation.ArrayAnnotation;
import ida.pelab.prt.pattern.LoopVariable;
import ida.pelab.prt.pattern.MatchedInstance;
import ida.pelab.prt.pattern.PatternInstanceParameter;
import ida.pelab.prt.pattern.PatternLevel;


public class CCodeGenerator implements ICodeGenerator {
	
	private CcodePlatform platform;
	 @Override
	 public  CGNAnnotation generateCode(MatchedInstance instance)
	 {
		 
		 	String annotationString="";
		 	CodeExtension codeExModule= new CodeExtension();
			String result="   "+instance.getInstance().getPattern().getOutputSyntax().getPrintingName().toLowerCase() +"(";
			ArrayList<PatternInstanceParameter> params= instance.getInstance().getParameters();
			ArrayList<LoopVariable> loopVariableList= new ArrayList<LoopVariable>();
						
			for(int i=0;i<params.size();i++)
			{
				PatternInstanceParameter cuInstace= params.get(i);
				if(cuInstace.isNegated())
					result+="-";
				if (cuInstace.isDivided())
					result+="1/";
				switch (cuInstace.getType()) {
				case Number:
						result+= params.get(i).getTreeNode().toString()+",";
					break;
				case String:
					result+= params.get(i).getStringValue()+ ",";
					break;
				case Identifier:

						result+= params.get(i).getTreeNode().toString()+",";		
					break;
				case Vector:
						result+= params.get(i).getTreeNode().toString()+",";
					break;				
				case Loopvariable:
					
					  LoopVariable tmpLV= (LoopVariable)params.get(i).getTreeNode();
					  loopVariableList.add(tmpLV);					   					 
					  annotationString+=codeExModule.addLoopVariable(instance.getTreeNodeReference(), tmpLV);
					  result+="rng_" +codeExModule.getNodeName(tmpLV.getVariableExpression())+ ",";
					break;
					
				case Subscripted:
					if(instance.getInstance().getPattern().getLevel().equals(PatternLevel.Level1))			
						{
							if(!isDuplicatedParameter(params, i))
							 annotationString+=codeExModule.addVectorStatement(params.get(i).getTreeNode()
									 ,ArrayAnnotation.toArrayBoundaryList((ArrayAccess)params.get(i).getTreeNode(), loopVariableList));
							result+="vtr_" +codeExModule.getNodeName(params.get(i).getTreeNode())+ ",";
						}
					else if(instance.getInstance().getPattern().getLevel().equals(PatternLevel.Level2))
						{
							if(!isDuplicatedParameter(params, i))							
								annotationString+=codeExModule.addMatrixStatement(params.get(i).getTreeNode()
										,ArrayAnnotation.toArrayBoundaryList((ArrayAccess)params.get(i).getTreeNode(), loopVariableList));
							result+="mtrx_" +codeExModule.getNodeName(params.get(i).getTreeNode())+ ",";
						}					
					break;
					
				case nonSubscripted:
					result+= params.get(i).getTreeNode().toString()+",";
				default:
					break;
				}
			}
		  if(result.endsWith(","))
				result=result.substring(0,result.length()-1);
			result+=");";		
		  
			
			annotationString+=result;
		 addExtraBlock(instance.getTreeNodeReference());
		return new CGNAnnotation(annotationString);
		 
	 }
	
	 private static void addExtraBlock(Traversable subTreeRoot)
	 {
		 CompoundStatement comp= new CompoundStatement();
		 subTreeRoot.getParent().getChildren().add(subTreeRoot.getParent().getChildren().indexOf(subTreeRoot), comp);
		 subTreeRoot.getParent().getChildren().remove(subTreeRoot);
		 comp.getChildren().add(subTreeRoot);
		 
	 }
	 private  static boolean isDuplicatedParameter(ArrayList<PatternInstanceParameter> paramList,int index)
	 {
		 boolean result=false;
		 for(int i=0; i<index;i++)
			 if(paramList.get(i).getTreeNode().equals(paramList.get(index).getTreeNode()))
				 result=true;
		 return result;
	 }

	@Override
	public void runExtraOperation(Traversable irTreeRoot) {
		if(irTreeRoot instanceof Program)
		{
			CcodePlatform platform=getPlatform();
			platform.addDirective(platform.getWrapperFileName());
			CodeExtension.addIncludeStatements(irTreeRoot, platform.getDirectiveList());
		}
		
	}

	public void setPlatform(CcodePlatform platform) {
		this.platform = platform;
	}

	public CcodePlatform getPlatform() {
		if(platform==null)
			platform=new SettingLoader().loadSetting();
		return platform;
	}

	@Override
	public boolean hideCode() {
		// TODO Auto-generated method stub
		return getPlatform().isShowCode();
	}

	@Override
	public PatternLevel getLevel() {
		// TODO Auto-generated method stub
		return PatternLevel.Level1;
	}

	

}
