package ida.pelab.loopdistributionTool;





import ida.pelab.prt.generalfunctions.IgnoredNodes;

import java.util.ArrayList;
import java.util.List;

import cetus.analysis.DependenceVector;
import cetus.analysis.DDGraph.Arc;
import cetus.hir.ForLoop;
import cetus.hir.Traversable;

public class DirectedGraph {
	
	
	private ArrayList<Traversable> vertices ; 
	private ArrayList<Edge>   edges ;
	
	public DirectedGraph()
	{
		vertices = new ArrayList<Traversable>();
		edges = new ArrayList<Edge>();
				
	}
	
	public void addEdge(Edge inputEdge)
	{
		if(edges.indexOf(inputEdge)<0 &&
		inputEdge.getSink()!=inputEdge.getSource()		
		
		)
			edges.add(inputEdge);
		
		
		//Assuming that we have n nested loop from 1 to n  and we are at nested level  1<j<n
		//the following condition must be checked in order to ignore the dependencies in levels j<j1<n at level j 
		// since they are important for us as we do the distribution in post order 
		//inputEdge.getSink()!=inputEdge.getSource()
		
		
		
	}
	
	public void addVertex(Traversable vertex)
	{
		
		if(!vertices.contains(vertex))
			vertices.add(vertex);
	}
	
	public ArrayList<Edge> getEdges()
	{
		return edges;
	}

	public ArrayList<Traversable> getVertices(){
		return vertices;
	}
	
	public ArrayList<Traversable> getSuccessors(Traversable vertex){
		
		ArrayList<Traversable> output = new ArrayList<Traversable>();
		
		if(vertices.indexOf(vertex)>-1)
		{
			
			for(int i = 0 ; i< edges.size();i++)
				if(edges.get(i).getSource()==vertex)
					output.add(edges.get(i).getSink());												
			
		}
		
		return output;
	}
	
	public void ConstructFromForLoop(ForLoop loop ,ArrayList<Arc> subscriptedDependencies ,ArrayList<Dependency> scalarDependencies)
	{
		//IgnoredNodes tobeIgnored = new IgnoredNodes();
		List<Traversable>  bodychildren = loop.getBody().getChildren();
		for(int i= 0 ; i<bodychildren.size();i++)
			if(!IgnoredNodes.isUninteresting(bodychildren.get(i)))
				addVertex(bodychildren.get(i));
		
		for(int i = 0 ; i<scalarDependencies.size();i++)
		{
		   Dependency dps =scalarDependencies.get(i);
		   Edge edge = new Edge(dps.getSourceStatement(),
				   dps.getSinkStatement());
				   
		   addEdge(edge);
			
			
		}
		
		for(int i = 0; i<subscriptedDependencies.size();i++)
		{
			Arc dependencyArc = subscriptedDependencies.get(i);
			
			if(LoopDistributor.isBackwardDependency(dependencyArc,loop.getBody()))
			/*|| dependencyArc.getSink().getArrayAccess().toString().equals(dependencyArc.getSource().getArrayAccess().toString())*/											
				{
					if(isDependencyVectorValid(dependencyArc.getDependenceVector()))
					
						{
							Edge forward = new Edge(LoopDistributor.getEquivalentNodeLevel(dependencyArc.getSourceStatement(),loop.getBody()) , LoopDistributor.getEquivalentNodeLevel(dependencyArc.getSinkStatement(),loop.getBody()));
							Edge backward = new Edge(LoopDistributor.getEquivalentNodeLevel(dependencyArc.getSinkStatement(),loop.getBody()) , LoopDistributor.getEquivalentNodeLevel(dependencyArc.getSourceStatement(),loop.getBody()));
							addEdge(forward);
							addEdge(backward);
						}
					else 
					   {
						Edge backward = new Edge(LoopDistributor.getEquivalentNodeLevel(dependencyArc.getSinkStatement(),loop.getBody()) , LoopDistributor.getEquivalentNodeLevel(dependencyArc.getSourceStatement(),loop.getBody()));
						addEdge(backward);
					   }
				}	
				
			
			else
			{
				Edge  edge = new Edge(LoopDistributor.getEquivalentNodeLevel(dependencyArc.getSourceStatement(),loop.getBody()) , LoopDistributor.getEquivalentNodeLevel(dependencyArc.getSinkStatement(),loop.getBody()));
				addEdge(edge);
			}
		
			
			
		}
		

		
	}
	
	

	//a dependence cannot exist if it has a direction vector whose leftmost non-
	//“=” component is not “<” because that would mean that the sink of the
	//dependence occurs before the source, which is impossible
	
	private boolean isDependencyVectorValid(DependenceVector  dpVector)
	{
           String vectorString=dpVector.toString();
 
           int counter = 0 ;
           while(counter<vectorString.length())
           {
        	   if(vectorString.charAt(counter)=='>')
        		   break ;
        	   else if(vectorString.charAt(counter)=='=')
        		   counter++ ;
        	   else if(vectorString.charAt(counter)=='<' || vectorString.charAt(counter)=='*')
        		   return true ;
        	   
           }
           
		return false ;
	}

}
