package ida.pelab.prt.analysis;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.LinkedList;





import cetus.hir.AccessExpression;

import cetus.hir.AccessOperator;
import cetus.hir.AnnotationStatement;

import cetus.hir.BreadthFirstIterator;
import cetus.hir.BreakStatement;
import cetus.hir.Case;
import cetus.hir.CompoundStatement;
import cetus.hir.ContinueStatement;
import cetus.hir.DeclarationStatement;
import cetus.hir.ExpressionStatement;
import cetus.hir.ForLoop;
import cetus.hir.FunctionCall;
import cetus.hir.IfStatement;
import cetus.hir.Label;
import cetus.hir.Loop;
import cetus.hir.NullStatement;
import cetus.hir.ReturnStatement;
import cetus.hir.Statement;
import cetus.hir.SwitchStatement;
import cetus.hir.Traversable;
import cetus.hir.Typecast;
import cetus.hir.UnaryExpression;
import cetus.hir.UnaryOperator;

import ida.pelab.prt.pattern.PatternLevel;
import ida.pelab.prt.pattern.PatternStructure;
import ida.pelab.prt.patternhierarchy.AnnotatedTree;


public class RecognitionAnalysis {

	
	private static ArrayList<String> invalidExpressions= new ArrayList<String>();
	private static ArrayList<ExpressionStatement> nonDetectedExpressions= new ArrayList<ExpressionStatement>();
	public static void Start()
	{
	 		setDetectedStatementCNT(0);
	 		setDetectedForloopCNT(0);
	 		setStatementCount(0);
	 		setLoopCount(0);
	 		setDetectedPatterns(0);
	 		setPointerClassExpression(0);
	 		setComputationalStatements(0);
	 		invalidExpressions.clear();
	 		nonDetectedExpressions.clear();
	 	
			
	}
	
	
	
	public static void updateAnalysisInfo(Traversable treeNode,PatternStructure newPattern)
	{
		if(!newPattern.isTrivial())
		{
			 setDetectedPatterns(getDetectedPatterns()+1);
			
			if(newPattern.getLevel().ordinal()> PatternLevel.Level0.ordinal())
			  {
				setDetectedForloopCNT(getDetectedForloopCNT()+1);
				// by our convension certain type of loops are considered invalid 
				// as a result they are not counted by the counter however if a pattern has been detected for these statements
				// we should update the loop count value
				if(isInvalid(treeNode))
					setLoopCount(getLoopCount()+1);
					
			  }
			
		}
		
	}
	
	public static void updateExpressionStatementInfo(Traversable treeNode)
	{
		if(treeNode instanceof ExpressionStatement)
		{
			setDetectedStatementCNT(getDetectedStatementCNT()+1);
			if(isInvalid((ExpressionStatement)treeNode))
				 setComputationalStatements(getComputationalStatements()+1);
			
			//if(checkforClassorPointers(treeNode))
				//setPointerClassExpression(getPointerClassExpression()-1);
		}
	}
	
	public static void updateNodeInfo(Traversable treeNode)
	{
	
		if(treeNode instanceof ExpressionStatement)
			{
			   // prunes the step, initialstatement and the condition of a for loop		
			   if(treeNode.getParent() instanceof CompoundStatement)
			   		{
				   		setStatementCount(getStatementCount()+1);
				   		if( !isInvalid((ExpressionStatement)treeNode))
						   {
					   		 setComputationalStatements(getComputationalStatements()+1);
					   		
					   		if(checkforClassorPointers(treeNode))
								setPointerClassExpression(getPointerClassExpression()+1);
						   }
				   		else invalidExpressions.add(treeNode.toString());
		   
			   		}
			   
			
			}
		else if(treeNode instanceof Statement && !( treeNode instanceof CompoundStatement))
		{
			setStatementCount(getStatementCount()+1);
			if(!isInvalid(treeNode))
			  {	
				 setComputationalStatements(getComputationalStatements()+1);
				if(checkforClassorPointers(treeNode))
						setPointerClassExpression(getPointerClassExpression()+1);
				   
				 if(treeNode instanceof ForLoop)
					   setLoopCount(getLoopCount()+1);
			  }
			else invalidExpressions.add(treeNode.toString());
		

		}
		
	}
	//represent the percentage of statements that contained pointers
	public static double getPointerCoverage()
	{
		
		if(computationalStatements==0)
			return 0;
		else 
			return (pointerClassExpression)/computationalStatements;
		
	}
	
	public static double getStatementCoverage()
	{
		
		if(computationalStatements==0)
			return 0;
		else 
			return (detectedStatementCNT+detectedForloopCNT)/computationalStatements;
		
	}
	
	
	public static double getForLoopCoverage()
	{
		
		if(loopCount==0)
			return 0;
		else 
			return detectedForloopCNT/loopCount;
		
	}
	
	// returns the ration valid computational to all statements
	public static double getValidCodeRatio()
	{
		
		if(statementCount==0)
			return 0;
		else

			return computationalStatements/statementCount;

		
	}
 
	public static void printAnalysisResult()
	{
		
			NumberFormat percentageFormat ;
		 	percentageFormat=NumberFormat.getInstance();
		 	percentageFormat.setMaximumFractionDigits(2);
		    System.out.println("[pattern recognizer] Number of Detected Patterns: "+ Integer.toString(getDetectedPatterns()) );
		    System.out.println("[pattern recognizer] Number of Statements: "+ Double.toString(getStatementCount()) );
		    System.out.println("[pattern recognizer] percentage of Detected Statements: "+ percentageFormat.format(getStatementCoverage()*100)+   "%" );
		    System.out.println("[pattern recognizer] " + Double.toString( detectedForloopCNT)+ " for loops out of " +
		    Double.toString( loopCount) + " (" +  percentageFormat.format(getForLoopCoverage()*100) +   "%) were recognized" 		
		    );
		    System.out.println();
		    
		    fetchNonDetectedStatements();
		    
		
		
		
	}
	
	
	private static void fetchNonDetectedStatements()
	{
		ArrayList<Statement> nonDetectedNodes=AnnotatedTree.getNonDetectedPattern();
				
		for(int i= 0; i< nonDetectedNodes.size();i++)
			if(nonDetectedNodes.get(i) instanceof ExpressionStatement && !nonDetectedNodes.get(i).toString().equals(""))
			{
				ExpressionStatement expStatement=((ExpressionStatement)nonDetectedNodes.get(i)).clone();
				//if(expStatement.getAnnotations()!=null)
					//expStatement.getAnnotations().clear();
				nonDetectedExpressions.add(expStatement);	
			}
			
 	
	}
	
	public static ArrayList<ExpressionStatement> getNonDetectedStatements()
	{
		return nonDetectedExpressions;	
	}
	
	private static void setStatementCount(double statementCount) {
		RecognitionAnalysis.statementCount = statementCount;
	}
	public static double getStatementCount() {
		return statementCount;
	}



	private static void setDetectedStatementCNT(double detectedStatementCNT) {
		RecognitionAnalysis.detectedStatementCNT = detectedStatementCNT;
	}





	public static double getDetectedStatementCNT() {
		return detectedStatementCNT;
	}





	private static void setDetectedForloopCNT(double detectedForloopCNT) {
		RecognitionAnalysis.detectedForloopCNT = detectedForloopCNT;
	}





	public  static double getDetectedForloopCNT() {
		return detectedForloopCNT;
	}





	private static void setLoopCount(double loopCount) {
		RecognitionAnalysis.loopCount = loopCount;
	}



	public static double getLoopCount() {
		return loopCount;
	}





	private static void setDetectedPatterns(int detectedPatterns) {
		RecognitionAnalysis.detectedPatterns = detectedPatterns;
	}



	public static int getDetectedPatterns() {
		return detectedPatterns;
	}


    private static boolean isInvalid(Traversable node)
    {
    	
        if( node instanceof AnnotationStatement|| node instanceof BreakStatement || node instanceof ContinueStatement 
    			  || node instanceof CompoundStatement  ||node instanceof  DeclarationStatement  || 
    			  node instanceof ReturnStatement|| node instanceof NullStatement || node instanceof Label || node instanceof Case)
    		  return true ;
    	
    	// a loop is Invalid if all the statements in the body are invalid
        else if(node instanceof Loop)
    	{
    		
    		Loop lp=(Loop)node;
    	    CompoundStatement body=(CompoundStatement)lp.getBody();
    	    return isInvalidBody(body);    
    		    		
    	}
        else if(node instanceof IfStatement)
        {
        	IfStatement ifstmt=(IfStatement)node ;
           
        	CompoundStatement thenbody=(CompoundStatement)ifstmt.getThenStatement();
        	if(!isInvalidBody(thenbody))
        		return false ;
        	 
        	 if(ifstmt.getElseStatement()!=null)
        	 {
        		 CompoundStatement elsebody=(CompoundStatement)ifstmt.getElseStatement();
        		 if(!isInvalidBody(elsebody))
             		return false ;	 
            	
            }
        	 
        	 return true;  	
        }
        else if(node instanceof SwitchStatement)
        {
        //	SwitchStatement sw=(SwitchStatement)node;
        	if(node.getChildren().size()>1)
        	{
        		CompoundStatement body =(CompoundStatement)node.getChildren().get(1);
        		return isInvalidBody(body);  
        	}
              	 return false ;
      	 
        }
    	
        return false ;
    }
    
    
    private static boolean isInvalidBody(CompoundStatement body)
    {
    	
    	for(int i= 0 ;i<body.getChildren().size();i++)
		{
		    Traversable child =body.getChildren().get(i);
		    if(child instanceof ExpressionStatement)
		    	{
		    		if( !isInvalid((ExpressionStatement)child))
		    			return false ;
		    	}
		    		
		  else  if(!isInvalid(child))
			 return false ;
			
		}
    	
    	return true;
    	
    	
    }
    public static boolean isInvalid(ExpressionStatement node)
    {
    	
 	   
    	if( node.getExpression() instanceof FunctionCall)
    	{
    		FunctionCall fnCall =(FunctionCall)node.getExpression();
     	    return  !isFunctionNameValid(fnCall.getName().toString());
    	
    	}
    	
    	return hasTypeCastExpression(node);
    	//return false ;
    	
    	
    	
    }

    private static boolean isFunctionNameValid(String name)
    {

    	
	 return	name.equals("exp") || name.equals("log") || name.equals("sqrt") || name.equals("cos") ||name.equals("sin") ||name.equals("tan") ||
		name.equals("twoto") || name.equals("pow") || name.equals("max" )|| name.equals("min" ) ;

    	/*return name.equals("printf") || name.equals("scanf") || name.equals("fprintf") ||
    	
    			name.equals("fscanf") || name.equals("fopen") || name.equals("fclose") ||
    			name.equals("malloc") || name.equals("free") || name.equals("exit" );*/
    	
    }

	public static void setPointerClassExpression(int pointerClassExpression) {
		RecognitionAnalysis.pointerClassExpression = pointerClassExpression;
	}


	// checks for the existence of pointers or accessing member of a class
   private static boolean checkforClassorPointers(Traversable treeNode)
   {
	   if(treeNode==null)
		   return false;
	   if(treeNode instanceof Loop)
		   return (checkforClassorPointers(((Loop)treeNode).getCondition()))   ;
	   if(treeNode instanceof IfStatement)
		   return (checkforClassorPointers(((IfStatement)treeNode).getControlExpression()))   ;
	   if(treeNode instanceof SwitchStatement)
		   return (checkforClassorPointers(((SwitchStatement)treeNode).getExpression()))   ;
   
	   BreadthFirstIterator<Traversable> bfs = new BreadthFirstIterator<Traversable>(treeNode);
	  
	   //covers a->b expressions
	   LinkedList<AccessExpression> accessExprList=bfs.getList (AccessExpression.class);
	   if(accessExprList.size()>0)
	   {
		   for(int i= 0 ; i< accessExprList.size();i++)
		   if(accessExprList.get(i).getOperator().equals(AccessOperator.POINTER_ACCESS))
		   	return true ;
	   }
	   bfs.reset();
	   LinkedList<UnaryExpression> unLinkList = bfs.getList(UnaryExpression.class);
	   //covers  (*a)
	   if(unLinkList.size()>0)	   
		   for(int i= 0 ;i< unLinkList.size(); i++)
		    if(unLinkList.get(i).getOperator()==UnaryOperator.DEREFERENCE)
		    	return true ;
	    
	 
	   return false ;
	   
	
	   
   }
	
	public static int getPointerClassExpression() {
		return pointerClassExpression;
	}




	  private static boolean hasTypeCastExpression(Traversable treeNode)
	   {
		   BreadthFirstIterator<Traversable> bfs = new BreadthFirstIterator<Traversable>(treeNode);
		  
		   //covers a->b expressions
		   if(bfs.getList(Typecast.class).size()>0)
			   return true ;
		
		   return false ;
		
		   
	   }





	public static void setComputationalStatements(double computationalStatements) {
		RecognitionAnalysis.computationalStatements = computationalStatements;
	}



	public static double getComputationalStatements() {
		return computationalStatements;
	}

	// represents the number of all Statements inside the Code
	private static double statementCount;
	//represents the set of valid statements for code
	private static double computationalStatements;
	// represents the number of for-loops in the code block
	private static double loopCount;
	// represents the number of detected patterns in Expression Statements type
	private static double detectedStatementCNT;
	// represents the number of detected Level1 and Level2 patterns
	private static double detectedForloopCNT;
	//represents the number of Level0 and higher level pattern
	private static int detectedPatterns;
	
	//represents the number of statements that contained pointers
	private static int pointerClassExpression;


	
}
