package ida.pelab.prt.annotation;

import ida.pelab.cgn.ICodeGenerator;
import ida.pelab.cgn.TextComment.CommentCodeGenerator;

import ida.pelab.prt.pattern.MatchedInstance;

import ida.pelab.prt.pattern.PatternLevel;
import ida.pelab.prt.patternhierarchy.AnnotatedTree;


import java.util.ArrayList;
import java.util.List;






import cetus.hir.Annotation;


import cetus.hir.Expression;
import cetus.hir.ExpressionStatement;
import cetus.hir.ForLoop;
import cetus.hir.FunctionCall;
import cetus.hir.IfStatement;


import cetus.hir.PostOrderIterator;
import cetus.hir.PragmaAnnotation;
import cetus.hir.Statement;

import cetus.hir.Traversable;

public class AnnotationModule {
	
	
	private static ICodeGenerator codeTool;
	// Annotates the node with the full list of detected patterns
	// suitable for debugging purposes
	public static void annotateAllChildren(Traversable  irTreeRoot)
	{
		
    	PostOrderIterator<Traversable> postOrderIt = new PostOrderIterator<Traversable>(irTreeRoot);
    	if(irTreeRoot instanceof ForLoop)
			removePragmaAnnotation((ForLoop)irTreeRoot);
	 	while(postOrderIt.hasNext())
		{
		  Traversable nextNode =(Traversable)postOrderIt.next();
		  ArrayList<MatchedInstance> matchedSubPattern= AnnotatedTree.GetChildrensPattern(nextNode);
		  for(int i=0 ; matchedSubPattern!=null && i< matchedSubPattern.size();i++)
			     {
				 	MatchedInstance tmpInstance = matchedSubPattern.get(i);
				 	if(tmpInstance.getInstance().getPattern().getLevel().ordinal()> PatternLevel.Trivial.ordinal())	 
					  addAnnotationtoIR(tmpInstance);			
				  
			    }
			
		}
		
		
	}

	
	public static void annotateSourceCode(Traversable  irTreeRoot)
	{
		getCodeTool().runExtraOperation(irTreeRoot);
		annotateParentNode(irTreeRoot, getCodeTool().hideCode());
	}
	
	/**
	 * Annotate each node with at most one pattern at the statement level.
	 * @param irTreeRoot
	 * @param showSourceCode
	 */
	public static void annotateParentNode(Traversable  irTreeRoot,boolean showSourceCode)
	{
		if(irTreeRoot instanceof ForLoop)
			removePragmaAnnotation((ForLoop)irTreeRoot);
		
		  ArrayList<MatchedInstance> matchedSubPattern= AnnotatedTree.GetChildrensPattern(irTreeRoot);
		  if(matchedSubPattern!=null && matchedSubPattern.size()>0)
		  {
			  
			  for(int i=0 ; i< matchedSubPattern.size();i++)
			  {
				  MatchedInstance tempInstance = matchedSubPattern.get(i);		
				  
				  if(tempInstance.getInstance().getPattern().getLevel().ordinal()>  getCodeTool().getLevel().ordinal()-1 &&
						!tempInstance.getInstance().getPattern().isHelpPattern()  )
				    {	 

					  addAnnotationtoIR(tempInstance);					 
					  if(!showSourceCode)
						  hideEquivalentCodeStatement(tempInstance);
					  
				     }
				  
				  
				  else if(tempInstance.getInstance().getPattern().getName().equals("No_Pattern"))
					  annotateParentNode(tempInstance.getTreeNodeReference(),showSourceCode);
				  else removeExtraAnnotation(tempInstance);

				  
			  }
			  
     	 }
		
	     else  
		  {
			  for(int i= 0; irTreeRoot.getChildren()!=null && i<irTreeRoot.getChildren().size();i++)
				  annotateParentNode(irTreeRoot.getChildren().get(i),showSourceCode);
			  
		  }  
		
	}
	
	
		
	
	private static void addAnnotationtoIR(MatchedInstance matchedInstance)
	{
			
		if(matchedInstance.getInstance().getPattern().getLevel().ordinal()< getCodeTool().getLevel().ordinal())
			return;
				Annotation annotation= getCodeTool().generateCode(matchedInstance); 
				Traversable equivalentNode = null;

				equivalentNode=matchedInstance.getTreeNodeReference();		
				
			
				if(equivalentNode instanceof ForLoop)				
					((ForLoop)equivalentNode).annotate(annotation);								
				else if(equivalentNode instanceof FunctionCall)
					((FunctionCall)equivalentNode).getStatement().annotate(annotation);
				else if(equivalentNode instanceof Expression)
					((Expression)equivalentNode).getStatement().annotate(annotation);
				else if(equivalentNode instanceof IfStatement)
					((IfStatement)equivalentNode).annotate(annotation);
				else if(equivalentNode instanceof ExpressionStatement)
					((ExpressionStatement)equivalentNode).annotate(annotation);
			
			//removing pragma Annotation
			removeExtraAnnotation(matchedInstance);
            if(matchedInstance.getInstance().getPattern().getLevel().equals(PatternLevel.Level2))
            	removeExtraAnnotation(AnnotatedTree.GetChildrensPattern(matchedInstance.getTreeNodeReference()).get(0));
		
		
	}
	
	// hide the source code for  the current statement by setting an empty print method
	private static  void hideEquivalentCodeStatement(MatchedInstance matchedInstance)
	{

	    	Traversable currentNode =matchedInstance.getTreeNodeReference();
	    	if(currentNode instanceof Statement)
	    		((Statement)currentNode).setPrintMethod(PrintingMethods.getSkipPrintMethod(Statement.class));
	    		
	    	else if(currentNode instanceof Expression && currentNode.getParent() instanceof ExpressionStatement)
	    	{
	    		
	    		((Expression)currentNode).setPrintMethod(PrintingMethods.getSkipPrintMethod(Expression.class));
					((Statement)currentNode.getParent()).setPrintMethod(PrintingMethods.getSkipPrintMethod(Statement.class));
	    		
	    		
	    	}	    							 

		
		
		
	}
	
	
	//Data Dependency Analysis generates pragma annotations which are not useful
	//in our case
private static void removePragmaAnnotation(ForLoop flp)
{

		List<Annotation> annotationList= (flp).getAnnotations();
		if(annotationList!=null)
		for(int i = 0 ; i<annotationList.size();i++)
			if( annotationList.get(i) instanceof PragmaAnnotation)
			{
				annotationList.remove(i);
				i--;
			}
}
	
	
// removing the pragma annotation	
private static void removeExtraAnnotation(MatchedInstance structure)
{

	
    	if(structure.getTreeNodeReference() instanceof ForLoop)
    	removePragmaAnnotation((ForLoop)structure.getTreeNodeReference());
    
}



public static void setCodeTool(ICodeGenerator codeTool) {
	AnnotationModule.codeTool = codeTool;
}



public static ICodeGenerator getCodeTool() {
	if(codeTool==null)
		codeTool= new CommentCodeGenerator();
	return codeTool;
}

public static boolean  targetArchitectureDefined(){
	return  codeTool!=null;
}
	
	

	

}
