package ida.pelab.prt.patternrecognitionTool;


import java.io.IOException;



import ida.pelab.cgn.XMLSettingReader;
import ida.pelab.cgn.TextComment.CommentCodeGenerator;
import ida.pelab.cgn.ccode.CCodeGenerator;
import ida.pelab.loopdistributionTool.DependencyPool;

import ida.pelab.loopdistributionTool.TopDownLoopDistributor;

import ida.pelab.prt.analysis.RecognitionAnalysis;
import ida.pelab.prt.annotation.AnnotationModule;



import ida.pelab.prt.pattern.SubscriptedIndexTree;
import ida.pelab.prt.patterndefinition.PatternParser;
import ida.pelab.prt.patternhierarchy.AnnotatedTree;


import ida.pelab.prt.temporalAnalysis.TemporalAnalysis;
import cetus.hir.Program;

public class PatternRecognitionTool {
	
	
	private Program cSourceCode ;
	public PatternRecognitionTool(Program source)
	{
		setcSourceCode(source);
	}
	public void setcSourceCode(Program cSourceCode) {
		this.cSourceCode = cSourceCode;
	}
	public Program getcSourceCode() {
		return cSourceCode;
	}

	

	
	/**
	 * runs the pattern recognition tool on the input file.
	 * @param showcode
	 */
	public void run()
	{
	
		System.out.println("[pattern recognizer] loading the patterns,please wait...");
		if(PatternParser.parse())
			{
				RecognitionAnalysis.Start();
				System.out.println("[pattern recognizer]applying the transformation tool ...");
				TemporalAnalysis.measureTransformTime();
				try{
					TopDownLoopDistributor.applyForLoopDistribution(cSourceCode);
					}
				catch(Exception ex)
					{
						System.err.println("error during loop distribution process...... loop distribution is skipped");
					}
				TemporalAnalysis.stopTransformTime();
				TemporalAnalysis.measureRecognitionTime();
				System.out.println("[pattern recognizer]applying the recognition tool ...");
				RecognitionAlgorithm.traverse(cSourceCode);		   
				System.out.println("[pattern recognizer] printing the detected patterns ...");
				chooseTargetArchtecture();
				AnnotationModule.annotateSourceCode(cSourceCode);

				try {
						cSourceCode.print();
					} catch (IOException e) {
						// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			TemporalAnalysis.stopRecognitionTime();	
		    RecognitionAnalysis.printAnalysisResult();
		    TemporalAnalysis.printResults();
		    AnnotatedTree.Clear();
		    DependencyPool.clear();
		    SubscriptedIndexTree.Clear();
		    
		    
		  
		   
		}
		else System.out.println("error loading the patterns");
		
		

		
	}
	
	public static void execute(String fileString)
	{		
		String [] programArgs =  new String[2]; 
		programArgs[0]="-idaprt=";						
		programArgs[1]=	 fileString;
		cetus.exec.Driver.main(programArgs);
		
	}
	
	public void chooseTargetArchtecture()
	{
	 if(!AnnotationModule.targetArchitectureDefined())
	 {
		 String defaultArch= new  XMLSettingReader().getDefaultPlatform();
		 if(defaultArch.equals("Annotation"))
			 AnnotationModule.setCodeTool(new CommentCodeGenerator());
		 else if(defaultArch.equals("Ccode"))
			 AnnotationModule.setCodeTool(new CCodeGenerator());
	 }
	}

}
