package ida.pelab.prt.patternrecognitionTool;

import ida.pelab.prt.constraints.Constraint;
import ida.pelab.prt.constraints.RuleParameterType;
import ida.pelab.prt.generalfunctions.LoopFunctions;
import ida.pelab.prt.pattern.MatchedInstance;
import ida.pelab.prt.pattern.PatternInstanceParameter;
import ida.pelab.prt.pattern.PatternLevel;
import ida.pelab.prt.pattern.PatternStructure;
import ida.pelab.prt.pattern.PatternStructureOperator;
import ida.pelab.prt.pattern.SubscriptedIndexTree;
import ida.pelab.prt.patternhandlers.MappedInstanceHandler;
import ida.pelab.prt.patternhierarchy.AnnotatedTree;
import ida.pelab.prt.patternhierarchy.PatternHierarchyGraph;



import java.util.ArrayList;

import cetus.hir.ArrayAccess;
import cetus.hir.AssignmentExpression;
import cetus.hir.AssignmentOperator;
import cetus.hir.BinaryExpression;
import cetus.hir.BinaryOperator;
import cetus.hir.FloatLiteral;
import cetus.hir.FunctionCall;
import cetus.hir.Identifier;
import cetus.hir.IfStatement;
import cetus.hir.IntegerLiteral;
import cetus.hir.UnaryExpression;
import cetus.hir.UnaryOperator;

import cetus.hir.ForLoop;

import cetus.hir.Traversable;


/**
 *
 */
public class StructuralComparison {
	
	
	
	/**
	 * compares the constraints of the pattern against the structure of the candidate sub-tree.
	 * @param pattern
	 * @param subtreeRoot
	 * @param patternList
	 * @return
	 */
	public static  boolean compare (PatternStructure pattern,Traversable subtreeRoot,ArrayList<MatchedInstance> patternList)
	{
		if(pattern.hasCustomMatchingRules())
			return AuxiliaryRecognitionModule.compare(pattern, subtreeRoot, patternList);
		
	
		if(pattern.isPruneIFCondtion())
			prundematchedInstancesinIFCond(subtreeRoot, patternList);
		
		return (mapOperator(pattern.getOperator(), subtreeRoot)) &&
			compareXMLStructure(pattern, patternList) &&
			checkPatternContraints(subtreeRoot, pattern.getConstraints(), patternList);
		
	}

	
/**
 * maps the operator specified for the pattern with the candidate sub-tree.
 * @param operator
 * @param node
 * @return
 */
private static boolean mapOperator(PatternStructureOperator operator,Traversable node)
{
		
	switch(operator){
	case Identifier:
		return node instanceof Identifier;
	case IntegerLiteral:
		return node instanceof IntegerLiteral;
	case FloatLiteral:
		return node instanceof FloatLiteral;
	case ArrayAccess:
		return node instanceof ArrayAccess;
	case FunctionCall:
		return node instanceof FunctionCall;	
	case AND:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.LOGICAL_AND);
	case OR:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.LOGICAL_OR);
	case MOD:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.MODULUS);
	case BLT:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.COMPARE_LT);
	case BGT:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.COMPARE_GT);
	case BGE:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.COMPARE_GE);
	case BLE:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.COMPARE_LE);
	case BNE:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.COMPARE_NE);
	case BEQ:
		return node instanceof BinaryExpression&& ((BinaryExpression)node).getOperator().equals(BinaryOperator.COMPARE_EQ);
	case ASGN:
		return node instanceof AssignmentExpression && ((AssignmentExpression)node).getOperator().equals(AssignmentOperator.NORMAL);
	case ADD:
		return node instanceof BinaryExpression && ((BinaryExpression)node).getOperator().equals(BinaryOperator.ADD);	
	case SUB:
		return node instanceof BinaryExpression && ((BinaryExpression)node).getOperator().equals(BinaryOperator.SUBTRACT);		
	case MUL:
		return node instanceof BinaryExpression && ((BinaryExpression)node).getOperator().equals(BinaryOperator.MULTIPLY);		
	case DIV:
		return node instanceof BinaryExpression && ((BinaryExpression)node).getOperator().equals(BinaryOperator.DIVIDE);		
	case ADDEQ:
		return node instanceof AssignmentExpression && ((AssignmentExpression)node).getOperator().equals(AssignmentOperator.ADD);
	case SUBEQ:
		return node instanceof AssignmentExpression && ((AssignmentExpression)node).getOperator().equals(AssignmentOperator.SUBTRACT);
	case MULEQ:
		return node instanceof AssignmentExpression && ((AssignmentExpression)node).getOperator().equals(AssignmentOperator.MULTIPLY);			
	case DIVEQ:
		return node instanceof AssignmentExpression && ((AssignmentExpression)node).getOperator().equals(AssignmentOperator.DIVIDE);					
	case ForLoop:
		return node instanceof ForLoop &&  LoopFunctions.isForLoopValid((ForLoop)node);
	case IFStatement:
		return node instanceof IfStatement;
	case PREINC:
		return	node instanceof UnaryExpression && ((UnaryExpression)node).getOperator().equals(UnaryOperator.PRE_INCREMENT);
	case PREDEC:
		return	node instanceof UnaryExpression && ((UnaryExpression)node).getOperator().equals(UnaryOperator.PRE_DECREMENT);
	case POSTINC:
		return	node instanceof UnaryExpression && ((UnaryExpression)node).getOperator().equals(UnaryOperator.POST_INCREMENT);
	case POSTDEC:
		return	node instanceof UnaryExpression && ((UnaryExpression)node).getOperator().equals(UnaryOperator.POST_DECREMENT);
	case ADRSS:
		return	node instanceof UnaryExpression && ((UnaryExpression)node).getOperator().equals(UnaryOperator.ADDRESS_OF);
	case PONTR:
		return	node instanceof UnaryExpression && ((UnaryExpression)node).getOperator().equals(UnaryOperator.DEREFERENCE);	
	default:
			return false;
			
	}
		
}
	
	
private static boolean compareXMLStructure(PatternStructure pattern, ArrayList<MatchedInstance> patternList)
{
		if(patternList!=null)
		 {			
		   if( pattern.getStructureElement().size()!=patternList.size()) return false ;
		   
		   else for(int i=0; i<patternList.size();i++)
			{
			  int j;
			  ArrayList<String> diffVariations =pattern.getStructureElement().get(i);
			  for( j=0;j< diffVariations.size();j++)			  
				 if(compareVariationName( diffVariations.get(j), patternList.get(i).getInstance().getPattern().getName()))
				  break;
			  
				if(j==diffVariations.size())
						return false;								
			}	
	   }		
   return true;
	
}


private static boolean compareVariationName(String variationName,String patternSubStructure)
{  
	if(variationName.equals("ALLLEVEL0"))
		return PatternHierarchyGraph.hasPatternatLevel(patternSubStructure, PatternLevel.Level0);
	 else if(variationName.equals("ALLLEVEL1"))
		return PatternHierarchyGraph.hasPatternatLevel(patternSubStructure, PatternLevel.Level1);
	 else if(variationName.equals("ALLLEVEL2"))
		return PatternHierarchyGraph.hasPatternatLevel(patternSubStructure, PatternLevel.Level1);
	else return  variationName.equals(patternSubStructure);
}

private static boolean checkPatternContraints(Traversable rootnode, ArrayList<Constraint> constraints, ArrayList<MatchedInstance> patternList)
{
	
	for(int i=0 ; i< constraints.size();i++)
		if(!constraints.get(i).check(rootnode,patternList))
			return false;
	return true;
	
	
}



public static Traversable mapParameter(String referenceString,ArrayList<MatchedInstance> patternList )
{

	int indexPosition =-1;
	String [] values=referenceString.split("\\.");
	if(values.length<0 || values.length >4)
		return null;
	 for(int i=0; i<values.length;i++)
		 if(values[i].equals("index"))
			 indexPosition=i;
	if(indexPosition==-1)
		indexPosition=values.length;
	
	int [] digitArray = {-1,-1,-1};
	for(int i=0; i<indexPosition;i++)
		digitArray[i]=Integer.parseInt(values[i])-1;
	if(indexPosition!=values.length)
		digitArray[2]=Integer.parseInt(values[indexPosition+1])-1;
	
	Traversable outputResult=null;
	if(patternList.size()>digitArray[0])
		outputResult=patternList.get(digitArray[0]).getTreeNodeReference();
	if(digitArray[1]>-1)
		outputResult=patternList.get(digitArray[0]).getInstance().getParameters().get(digitArray[1]).getTreeNode();
	if(digitArray[2]>-1)
		if( !(outputResult instanceof ArrayAccess && ((ArrayAccess)outputResult).getIndices().size()> digitArray[2]) )
			outputResult=null;
		else 
			outputResult= ((ArrayAccess)outputResult).getIndex(digitArray[2]);
	

	return outputResult;
		
	
}


public static PatternInstanceParameter mapParameterInstance(PatternStructure pattern, String referenceString,ArrayList<MatchedInstance> patternList )
{
	 int iindex= referenceString.indexOf("index");
	if(iindex>-1)
	{
		Traversable arrayNode=mapParameter(referenceString.substring(0,iindex), patternList);
		ArrayList<MatchedInstance> ptrList= AnnotatedTree.GetChildrensPattern(arrayNode);
		return mapParameterInstance(pattern, referenceString.substring(iindex+6), ptrList);
		
	}
	String [] values=referenceString.split("\\.");
	int firstDigit=-1,secondDigit=-1;
	if(values.length<0 || values.length >2)
		return null;
	firstDigit= Integer.parseInt(values[0]);
	if(values.length==2)
		secondDigit=Integer.parseInt(values[1]);
	firstDigit--;
	secondDigit--;
	
	RuleParameterType type = pattern.getType(referenceString);
	if(patternList.size()>firstDigit)
	{
		if(secondDigit<0)
		{
			if(type.equals(RuleParameterType.Invalid))
				type=StructuralComparison.mapType(patternList.get(firstDigit).getTreeNodeReference());
			Traversable node =patternList.get(firstDigit).getTreeNodeReference();
			saveSubscriptedIndex(type, node);
			return  new PatternInstanceParameter(node,type);
		}
			
		else if(patternList.get(firstDigit).getInstance().getParameters().size()> secondDigit)
		{
			PatternInstanceParameter newPar= new PatternInstanceParameter(patternList.get(firstDigit).getInstance().getParameters().get(secondDigit));
			if(!type.equals(RuleParameterType.Invalid))
				newPar.setType(type);
			saveSubscriptedIndex(newPar.getType(), newPar.getTreeNode());
			return newPar;
		}
	}
	return   null;
		
		
	
}


private static void saveSubscriptedIndex(RuleParameterType type,Traversable node)
{  
	if(type.equals(RuleParameterType.Subscripted))
		SubscriptedIndexTree.addCandidateasIndex(node);
	
}






public static ArrayList<PatternInstanceParameter> mapAllParameterInstance(PatternStructure pattern, String referenceString,ArrayList<MatchedInstance> patternList )
{
	ArrayList<PatternInstanceParameter> resultList = new ArrayList<PatternInstanceParameter>();
	String [] values=referenceString.split("\\.");
	int firstDigit=-1,secondDigit=-1;
	if(values.length<0 || values.length >2)
		return null;
	firstDigit= Integer.parseInt(values[0]);
	if(values.length==2)
		secondDigit=Integer.parseInt(values[1]);
	firstDigit--;
	secondDigit--;
	
	RuleParameterType type ;
	if(patternList.size()>firstDigit)
	{
		if(patternList.get(firstDigit).getInstance().getParameters().size()> secondDigit)
		{
			for(int i=secondDigit; i< patternList.get(firstDigit).getInstance().getParameters().size();i++)
			{
				type= pattern.getType(Integer.toString(firstDigit+1)+"."+ Integer.toString(i+1));
				PatternInstanceParameter newPar= new PatternInstanceParameter(patternList.get(firstDigit).getInstance().getParameters().get(i));
				if(!type.equals(RuleParameterType.Invalid))
					newPar.setType(type);
				resultList.add(newPar);
			}
		}
	}
		
	return resultList;	
	
}











/**
 * maps the type of a Traversable class node to one of the defined types in the RuleParameterType enumeration.
 * @param node
 * @return
 */
public static RuleParameterType mapType( Traversable node)
{
	  if(node instanceof IntegerLiteral || node instanceof FloatLiteral)
		  return RuleParameterType.Number;
	  else if(node instanceof ArrayAccess)
		  return RuleParameterType.Vector;
	  else if(node instanceof Identifier)
		  return RuleParameterType.Identifier;
	  
	  return RuleParameterType.Invalid;
	
}


/**
 * Checks a certain sub-tree against all the horizontal patterns.
 * @param pattern
 * @param subtreeRoot
 * @param patternList
 * @return
 */
public static  boolean compareHorizontalStructure (PatternStructure pattern,Traversable subtreeRoot,ArrayList<MatchedInstance> patternList)
{
	if(pattern.hasCustomMatchingRules())
		return AuxiliaryRecognitionModule.compare(pattern, subtreeRoot, patternList);
	
	if(pattern.getStructureElement().size()> patternList.size())
		return false;
	ArrayList<MatchedInstance>  specifiedList=getPatternList(patternList,pattern);
	if(specifiedList.size()<pattern.getStructureElement().size())
		return false ;
	
	ArrayList<MatchedInstance> mappedInstances = new ArrayList<MatchedInstance>(pattern.getStructureElement().size());
	int []indexArray= new int [pattern.getStructureElement().size()];
	
	for(int i=0; i< indexArray.length;i++)
		indexArray[i]=0;
	
	int i=0;
	while(indexArray[0]< specifiedList.size())
	{
			for(;indexArray[i]<specifiedList.size() && 
			      ! isInstanceFound(specifiedList.get(indexArray[i]),pattern.getStructureElement().get(i)) ;indexArray[i]++ );
				if(indexArray[i]<specifiedList.size())
				{
					mappedInstances.add(specifiedList.get(indexArray[i]));
				  if(mappedInstances.size()==pattern.getStructureElement().size())
					{
						if(checkHorizontalContraints(mappedInstances, pattern.getConstraints()))
						{
							MappedInstanceHandler.handleHorizontal(pattern, subtreeRoot, mappedInstances);
							specifiedList.removeAll(mappedInstances);
							mappedInstances.clear();
							i=0;
							//indexArray[i]=0;
						}
						else
						{
							mappedInstances.remove(i);
							indexArray[i]++;
						}
													
					}
					else indexArray[i+1]=indexArray[i++]+1;
								
				}
			else
				{	
				 if(i>0)
					mappedInstances.remove(--i);
					indexArray[i]++;
				}
		
	}
		
return false ;
		

}

private static void prundematchedInstancesinIFCond(Traversable subTree,ArrayList<MatchedInstance> patternList)
{
	if(!(subTree instanceof IfStatement))
		return;
	IfStatement ifstmt=(IfStatement) subTree;
	Traversable condition =ifstmt.getControlExpression();
	
	for(int i=0;i< patternList.size();i++)
		if(patternList.get(i).getTreeNodeReference().equals(condition))
			patternList.remove(i--);
			
			
	
	
}


private static boolean isInstanceFound(MatchedInstance instance,ArrayList<String> subStructure)
{	
	 for(int i=0; i< subStructure.size();i++)
	   if(instance.getPatternName().equals(subStructure.get(i)))
		   return true;
	  
	 return false;
	
}

private static ArrayList<MatchedInstance> getPatternList(ArrayList<MatchedInstance>  patternList,PatternStructure pattern)
{
	

	
	ArrayList<MatchedInstance> resultList = new ArrayList<MatchedInstance>();
	for(int i= 0; i<patternList.size();i++)
		{
			for(int j=0;j<pattern.getStructureElement().size();j++)
				{
					boolean matched=false;
					for(int k=0; k<pattern.getStructureElement().get(j).size();k++)
						if(patternList.get(i).getPatternName().equals(pattern.getStructureElement().get(j).get(k)))						
							{
									resultList.add(patternList.get(i));
									matched=true;
									break;
							}
					if(matched)
						break;
				}			
						
		} 		
	return resultList ;
	
}


private static boolean checkHorizontalContraints(ArrayList<MatchedInstance>  mappedInstances, ArrayList<Constraint> constraints)
{
	for(int i=0; i< constraints.size();i++)
	   if(!constraints.get(i).check(null, mappedInstances))
		   	return false;
	
	return true;
	
}







}