package report;



import ida.pelab.prt.analysis.RecognitionAnalysis;

import java.io.File;
import java.io.FileWriter;

import java.io.PrintWriter;
import java.text.NumberFormat;
import java.util.ArrayList;


import cetus.hir.ExpressionStatement;


import dsptester.TestcaseRunner;




public class ReportGenerator {

	
public static void generateReport(ArrayList<ReportData> values) 
{
	
	 String outputFile =System.getProperty("user.dir") +"/" +"Results.html";
	 PrintWriter writer=null;
	 try {
	 
		    writer= new PrintWriter(new FileWriter(outputFile));
		    writer.println("<html><script src=\"sorttable.js\" type=\"text/javascript\"></script> <body>");
		    writer.println("<TABLE   class=\"sortable\" border=\"1\" bordercolor=\"075B01\" style=\"background-color:F7F7F7\"  cellpadding=\"3\" cellspacing=\"3\"><tr><td colspan=1></td><td>File</td><td>Size</td><td>#Statement</td><td>#Matched Patterns </td><td>Relevant Code Ratio</td><td>Statement Coverage</td><td>Pointer Occurrence Percentage</td><td>Matched For-Loops</td><td>Matched For-Loops Percentage</td><td>Transformation-Time</td><td>Recognition-Time</td> <td>Non Detected Level0</td></tr>");
		    for (int i = 0; i < values.size(); i++) {
    	  
		    	writer.println("<tr>");
		    	addColumn(writer,Integer.toString( i+1));
		    	addColumn(writer,"<A href=\"cetus_output/" +values.get(i).getFileName()+ "\">"+values.get(i).getFileName() +"</A>"); 
		    	addColumn(writer, values.get(i).getSize());
		    	if(values.get(i).getErrorStatement()!="")
		    		{ addColumn(writer, values.get(i).getErrorStatement());
		    		  for(int j= 0; j<9 ;j++)
		    			addColumn(writer, "");
		    		}
		    	else
		    	{
		    		addColumn(writer, values.get(i).getStatementCount());
		    		addColumn(writer, values.get(i).getDetectedPattern());
		    		addColumn(writer, values.get(i).getValidCodePercentage());
		    		addColumn(writer, values.get(i).getDetectedStatmentPercentage());

		    		addColumn(writer, values.get(i).getDetectedPointersInStatements());
		    		addColumn(writer, values.get(i).getDetectedForloop());
		    		addColumn(writer, values.get(i).getDetectedForloopPercentage());
		    		//addColumn(writer, values.get(i).getPointerFreeForLoopCoverage());
		    		addColumn(writer, values.get(i).getTransformationTime());
		    		addColumn(writer, values.get(i).getRecogntionTime());    	
		    		String nondetectedLink ="<A href=\"non_detected/" +values.get(i).getFileName()+"_nondetected.html"+ "\">"+values.get(i).getFileName()+"_nondetected.html" +"</A>"; 		    			
		    		addColumn(writer, nondetectedLink);
		    	}
    	  
		    	writer.println("</tr>");
		    	
		    	//  	int square = i * i;
		    	//pw.println("<TR><TD>" + i + "<TD>" + square);
		    }
		  writer.println("</TABLE>");
		  calculateAverage(writer);
		  writer.println("</body></html>");
		    
	 }
	catch (Exception e) {
		// TODO: handle exception
	}
	finally{
      writer.close();
	}

	
}


private static void  calculateAverage(PrintWriter pWriter)
{
   ArrayList<Double> loopCoverageInfo =TestcaseRunner.getForLoopDetectionInfo();
   ArrayList<Double> StatementCoverageInfo =TestcaseRunner.getStatementDetectionInfo();
   double sum = 0 ;
   for(int i= 0 ; i<loopCoverageInfo.size();i++)
	   sum+=loopCoverageInfo.get(i);
   NumberFormat percentageFormat ;
	percentageFormat=NumberFormat.getInstance();
	percentageFormat.setMaximumFractionDigits(2);
	pWriter.println("</br>"+ "Average ForLoop Coverage: "+ percentageFormat.format( sum/loopCoverageInfo.size()*100)+"%"+"</br>");
	sum= 0;
	  for(int i= 0 ; i<StatementCoverageInfo.size();i++)
		   sum+=StatementCoverageInfo.get(i);
	  pWriter.println( "Average Statement Coverage: "+ percentageFormat.format( sum/StatementCoverageInfo.size()*100)+"%"+ "</br>");
	   double stmtCov33 =countValue(StatementCoverageInfo,0, 33)/StatementCoverageInfo.size();
	   double stmtCov66 =countValue(StatementCoverageInfo,34, 66)/StatementCoverageInfo.size();
	   double stmtCov100 =countValue(StatementCoverageInfo,67, 101)/StatementCoverageInfo.size();
	   double forCov33 =countValue(loopCoverageInfo,0, 33)/loopCoverageInfo.size();
	   double forCov66 =countValue(loopCoverageInfo,34, 66)/loopCoverageInfo.size();
	   double forCov100 =countValue(loopCoverageInfo,67, 101)/loopCoverageInfo.size();
	   
	   pWriter.println("<TABLE border=1 ><tr><td colspan=1>Recognition Rate </td><td>0%-33%</td><td>34%-66%</td><td>67%-100%</td></tr>");  
	   pWriter.println("<tr><td >Statement Coverage</td><td>"+ percentageFormat.format(stmtCov33*100 ) +"%"+"</td><td>"+percentageFormat.format(stmtCov66*100 )+"%" +"</td><td>"+percentageFormat.format(stmtCov100*100 ) +"%"+"</td></tr>");
	   pWriter.println("<tr><td >For-Loop Coverage</td><td>"+ percentageFormat.format(forCov33*100 ) +"%"+"</td><td>"+percentageFormat.format(forCov66 *100) +"%"+"</td><td>"+percentageFormat.format(forCov100 *100)+"%" +"</td></tr>");
	   pWriter.println("</Table>");
	
}




public static void  printNonDetectedLevel0Patterns(String fileName,ArrayList<ExpressionStatement> statements)
{
    if(statements.size()==0)
    	return ;
	
	File  dir= new File(System.getProperty("user.dir") +"/non_detected/");
	if(!dir.exists())
		dir.mkdir();
	
	String outputFile =System.getProperty("user.dir") +"/non_detected/" + fileName+"_nondetected.html";
	 PrintWriter writer=null;
	 try {
	 
		
		 
		    writer= new PrintWriter(new FileWriter(outputFile));
		    writer.println("<html><script src=\"sorttable.js\" type=\"text/javascript\"></script> <body>");
		    writer.println("<TABLE border=1 ><tr><td colspan=1> Non Detected Level0 Statements </td></tr>");  
		    for(int i= 0 ; i<statements.size();i++)
		    {
		    	writer.println("<tr>");
		      if(!RecognitionAnalysis.isInvalid(statements.get(i)))
		      addColumn(writer, statements.get(i).toString());
		      writer.println("</tr>");
		    	
		    }
	 }
	 catch (Exception e) {
		// TODO: handle exception
		 e.printStackTrace();
	}
	 finally
	 {
		 writer.println("</Table></body></html>");
		 writer.close();
	 }
}

private static double countValue(ArrayList<Double> values,double lowerBound,double upperBound)
{
	double count = 0 ;
	for(int i= 0 ; i< values.size();i++)
		if(values.get(i)*100>=lowerBound && values.get(i)*100<upperBound)
			count++;
	return count;
	
}

private static void addColumn(PrintWriter writer,String value)
{
  writer.print("<td bordercolorlight=\"black\"  bordercolordark=\"black\"  style=\"font-weight: bold;\">");
  writer.print(value);
  writer.print("</td>");
}












}